import os
from typing import Callable, Optional, Union

import wandb
from flytekit import Secret
from flytekit.core.context_manager import FlyteContextManager
from flytekit.core.utils import ClassDecorator

WANDB_EXECUTION_TYPE_VALUE = "wandb-execution-id"
WANDB_CUSTOM_TYPE_VALUE = "wandb-custom-id"


class wandb_init(ClassDecorator):
    WANDB_PROJECT_KEY = "project"
    WANDB_ENTITY_KEY = "entity"
    WANDB_ID_KEY = "id"
    WANDB_HOST_KEY = "host"

    def __init__(
        self,
        task_function: Optional[Callable] = None,
        project: Optional[str] = None,
        entity: Optional[str] = None,
        secret: Optional[Union[Secret, Callable]] = None,
        id: Optional[str] = None,
        host: str = "https://wandb.ai",
        api_host: str = "https://api.wandb.ai",
        **init_kwargs: dict,
    ):
        """Weights and Biases plugin.
        Args:
            task_function (function, optional): The user function to be decorated. Defaults to None.
            project (str): The name of the project where you're sending the new run. (Required)
            entity (str): An entity is a username or team name where you're sending runs. (Required)
            secret (Secret or Callable): Secret with your `WANDB_API_KEY` or a callable that returns the API key.
                The callable takes no arguments and returns a string. (Required)
            id (str, optional): A unique id for this wandb run.
            host (str, optional): URL to your wandb service. The default is "https://wandb.ai".
            api_host (str, optional): URL to your API Host, The default is "https://api.wandb.ai".
            **init_kwargs (dict): The rest of the arguments are passed directly to `wandb.init`. Please see
                [the `wandb.init` docs](https://docs.wandb.ai/ref/python/init) for details.
        """
        if project is None:
            raise ValueError("project must be set")
        if entity is None:
            raise ValueError("entity must be set")
        if secret is None:
            raise ValueError("secret must be set")

        self.project = project
        self.entity = entity
        self.id = id
        self.init_kwargs = init_kwargs
        self.secret = secret
        self.host = host
        self.api_host = api_host

        # All kwargs need to be passed up so that the function wrapping works for both
        # `@wandb_init` and `@wandb_init(...)`
        super().__init__(
            task_function,
            project=project,
            entity=entity,
            secret=secret,
            id=id,
            host=host,
            api_host=api_host,
            **init_kwargs,
        )

    def execute(self, *args, **kwargs):
        ctx = FlyteContextManager.current_context()
        is_local_execution = ctx.execution_state.is_local_execution()

        if is_local_execution:
            # For location execution, always use the id. If `self.id` is `None`, wandb
            # will generate it's own id.
            wand_id = self.id
        else:
            if isinstance(self.secret, Secret):
                # Set secret for remote execution
                secrets = ctx.user_space_params.secrets
                wandb_api_key = secrets.get(key=self.secret.key, group=self.secret.group)
            else:
                # Get API key with callable
                wandb_api_key = self.secret()

            wandb.login(key=wandb_api_key, host=self.api_host)

            if self.id is None:
                # The HOSTNAME is set to {.executionName}-{.nodeID}-{.taskRetryAttempt}
                # If HOSTNAME is not defined, use the execution name as a fallback
                wand_id = os.environ.get("HOSTNAME", ctx.user_space_params.execution_id.name)
            else:
                wand_id = self.id

        run = wandb.init(project=self.project, entity=self.entity, id=wand_id, **self.init_kwargs)

        # If FLYTE_EXECUTION_URL is defined, inject it into wandb to link back to the execution.
        execution_url = os.getenv("FLYTE_EXECUTION_URL")
        if execution_url is not None:
            notes_list = [f"[Execution URL]({execution_url})"]
            if run.notes:
                notes_list.append(run.notes)
            run.notes = os.linesep.join(notes_list)

        output = self.task_function(*args, **kwargs)
        wandb.finish()
        return output

    def get_extra_config(self):
        extra_config = {
            self.WANDB_PROJECT_KEY: self.project,
            self.WANDB_ENTITY_KEY: self.entity,
            self.WANDB_HOST_KEY: self.host,
        }

        if self.id is None:
            wandb_value = WANDB_EXECUTION_TYPE_VALUE
        else:
            wandb_value = WANDB_CUSTOM_TYPE_VALUE
            extra_config[self.WANDB_ID_KEY] = self.id

        extra_config[self.LINK_TYPE_KEY] = wandb_value
        return extra_config



import os
from unittest.mock import Mock, patch

import pytest
from flytekitplugins.wandb import wandb_init
from flytekitplugins.wandb.tracking import WANDB_CUSTOM_TYPE_VALUE, WANDB_EXECUTION_TYPE_VALUE

from flytekit import Secret, task

secret = Secret(key="abc", group="xyz")


@pytest.mark.parametrize("id", [None, "abc123"])
def test_wandb_extra_config(id):
    wandb_decorator = wandb_init(
        project="abc",
        entity="xyz",
        secret=secret,
        id=id,
        host="https://my_org.wandb.org",
    )

    assert wandb_decorator.secret is secret
    extra_config = wandb_decorator.get_extra_config()

    if id is None:
        assert extra_config[wandb_decorator.LINK_TYPE_KEY] == WANDB_EXECUTION_TYPE_VALUE
        assert wandb_decorator.WANDB_ID_KEY not in extra_config
    else:
        assert extra_config[wandb_decorator.LINK_TYPE_KEY] == WANDB_CUSTOM_TYPE_VALUE
        assert extra_config[wandb_decorator.WANDB_ID_KEY] == id
    assert extra_config[wandb_decorator.WANDB_HOST_KEY] == "https://my_org.wandb.org"


@task
@wandb_init(project="abc", entity="xyz", secret=secret, tags=["my_tag"])
def train_model():
    pass


@patch("flytekitplugins.wandb.tracking.wandb")
def test_local_execution(wandb_mock):
    train_model()

    wandb_mock.init.assert_called_with(project="abc", entity="xyz", id=None, tags=["my_tag"])


@task
@wandb_init(project="abc", entity="xyz", secret=secret, tags=["my_tag"], id="1234")
def train_model_with_id():
    pass


@patch("flytekitplugins.wandb.tracking.wandb")
def test_local_execution_with_id(wandb_mock):
    train_model_with_id()

    wandb_mock.init.assert_called_with(project="abc", entity="xyz", id="1234", tags=["my_tag"])


@patch("flytekitplugins.wandb.tracking.FlyteContextManager")
@patch("flytekitplugins.wandb.tracking.wandb")
def test_non_local_execution(wandb_mock, manager_mock, monkeypatch):
    # Pretend that the execution is remote
    ctx_mock = Mock()
    ctx_mock.execution_state.is_local_execution.return_value = False

    ctx_mock.user_space_params.secrets.get.return_value = "this_is_the_secret"
    ctx_mock.user_space_params.execution_id.name = "my_execution_id"

    manager_mock.current_context.return_value = ctx_mock
    execution_url = "http://execution_url.com/afsdfsafafasdfs"
    monkeypatch.setattr("flytekitplugins.wandb.tracking.os.environ", {"FLYTE_EXECUTION_URL": execution_url})

    run_mock = Mock()
    run_mock.notes = ""
    wandb_mock.init.return_value = run_mock

    train_model()

    wandb_mock.init.assert_called_with(project="abc", entity="xyz", id="my_execution_id", tags=["my_tag"])
    ctx_mock.user_space_params.secrets.get.assert_called_with(key="abc", group="xyz")
    wandb_mock.login.assert_called_with(key="this_is_the_secret", host="https://api.wandb.ai")
    assert run_mock.notes == f"[Execution URL]({execution_url})"


def test_errors():
    with pytest.raises(ValueError, match="project must be set"):
        wandb_init()

    with pytest.raises(ValueError, match="entity must be set"):
        wandb_init(project="abc")

    with pytest.raises(ValueError, match="secret must be set"):
        wandb_init(project="abc", entity="xyz")


def get_secret():
    return "my-wandb-api-key"


@task
@wandb_init(project="my_project", entity="my_entity", secret=get_secret, tags=["my_tag"], id="1234")
def train_model_with_id_callable_secret():
    pass


@patch("flytekitplugins.wandb.tracking.os")
@patch("flytekitplugins.wandb.tracking.FlyteContextManager")
@patch("flytekitplugins.wandb.tracking.wandb")
def test_secret_callable_remote(wandb_mock, manager_mock, os_mock):
    # Pretend that the execution is remote
    ctx_mock = Mock()
    ctx_mock.execution_state.is_local_execution.return_value = False

    manager_mock.current_context.return_value = ctx_mock
    os_mock.environ = {}

    train_model_with_id_callable_secret()

    wandb_mock.init.assert_called_with(project="my_project", entity="my_entity", id="1234", tags=["my_tag"])
    wandb_mock.login.assert_called_with(key=get_secret(), host="https://api.wandb.ai")
